#include "jbsql.h"

char *command=NULL;
int nlines=1;
ColumnDesc *cols=NULL;
int ncols=0,nrows=0;
int recordlength=0;
DWORD tid=0;
CommandHistory *old_commands=NULL;
int nold_commands=0;
int current_command=0;
int allow_blank_lines=0;
char lastchar='\n';
char *last_object_type=NULL;
char *last_object_name=NULL;
int max_command_size=0,cur_command_size=0;
char *(old_tokens[NTOKENS])={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
int last_known_stmt=0;

void FreeCommand(void)
{
  free(command);
  max_command_size=0;
  cur_command_size=0;
}

void *MallocCommand(int size)
{
  void *command;

  command=malloc(size*2);
  max_command_size=size*2;
  cur_command_size=size;
  return command;
}

char *StrdupCommand(char *str)
{
  char *command;
  int size=strlen(str)+1;

  command=(char*)malloc(size*2);
  max_command_size=size*2;
  cur_command_size=size;
  strcpy(command,str);
  return command;
}

void ReallocCommand(int size)
{
  if (max_command_size<cur_command_size+size)
  {
    while (max_command_size<cur_command_size+size)
      max_command_size<<=1;
    command=(char*)realloc(command,max_command_size);
  }
  cur_command_size+=size;
}

char *Trim(char *str)
{
  int x;
 
  while (strchr(" \t\n\r",str[0]))
    memmove(&str[0],&str[1],strlen(str));
  for (x=strlen(str)-1;x>=0 && strchr(" \t\n\r",str[x]);x--)
    str[x]=0;
  return str;
}

void ToLower(char *str)
{
  while (*str)
  {
    *str=tolower(*str);
    str++;
  }
}

void ToUpper(char *str)
{
  while (*str)
  {
    *str=toupper(*str);
    str++;
  }
}

void StartTimer(TextWidget *text)
{
  text->start_time=GetTickCount();
}

char *StopTimer(TextWidget *text)
{
  DWORD total_time;
  static char str[100];
  DWORD days,hours,minutes,seconds,mseconds;

  total_time=GetTickCount()-text->start_time;
  if (INT_OPTION(SOUND))
  {
    if (total_time/1000>=INT_OPTION(SOUNDTIME))
      PlaySound(STR_OPTION(SOUNDFILE),NULL,SND_FILENAME|SND_ASYNC);
  }
  days=total_time/(24*60*60*1000);
  total_time%=24*60*60*1000;
  hours=total_time/(60*60*1000);
  total_time%=60*60*1000;
  minutes=total_time/(60*1000);
  total_time%=60*1000;
  seconds=total_time/1000;
  total_time%=1000;
  mseconds=total_time;

  if (!INT_OPTION(TIMING))
    strcpy(str,"");
  else
  {
    if (days)
      sprintf(str,"(%lu d %lu h %lu m %lu s %lu ms)",days,hours,minutes,seconds,mseconds);
    else if (hours)
      sprintf(str,"(%lu h %lu m %lu s %lu ms)",hours,minutes,seconds,mseconds);
    else if (minutes)
      sprintf(str,"(%lu m %lu s %lu ms)",minutes,seconds,mseconds);
    else if (seconds)
      sprintf(str,"(%lu s %lu ms)",seconds,mseconds);
    else
      sprintf(str,"(%lu ms)",mseconds);
  }
  return str;
}

int ReplaceString(char *str,char *srch,char *repl)
{
  size_t x;
  int replaced=0;
  
  for (x=0;x<strlen(str);x++)
  {
    if (x>strlen(str)-strlen(srch))
    {
      break;
    }
    if (memcmp(&str[x],srch,strlen(srch))==0)
    {
      memmove(&str[x],&str[x+strlen(srch)],strlen(str)-x-strlen(srch)+1);
      memmove(&str[x+strlen(repl)],&str[x],strlen(str)-x+1);
      memcpy(&str[x],repl,strlen(repl));
      x+=strlen(repl)-1;
      replaced=1;
    }
  }
  return replaced;
}

void RecallCommand(TextWidget *text,CommandHistory *old_command)
{
  int x,nl=0,lastx=0,slen;
  char *str;

  for (x=0;x<(int)strlen(command);x++)
  {
    if (command[x]=='\n')
      nl++;
  }
  RemoveLines(text,nl);
  WPrintf("\r");
  ClearCommand(text,0,0);
  text->no_process=1;
  if (old_command!=NULL)
  {
    str=strdup(old_command->command);
    slen=strlen(str);
    for (x=0;x<(int)slen;x++)
    {
      switch (str[x])
      {
        case '\n':
          str[x]='\0';
          InsertUnparsedLineBuffer(text,&str[lastx]);
          InsertNewLineBuffer(text);
          lastx=x+1;
          break;
        default:
          break;
      }
    }
    if (lastx<slen)
      InsertUnparsedLineBuffer(text,&str[lastx]);
    free(str);
    if (old_command->lastchar==';')
      InsertLineBuffer(text,"%c",old_command->lastchar);
    else if (old_command->lastchar=='/')
      InsertLineBuffer(text,"\n%c",old_command->lastchar);
  }
  text->no_process=0;
}

void PreviousStatement(TextWidget *text)
{
  int x,pos=-1;
  char lastchar;

  if (tid)
    return;
  if (current_command>0)
  {
    if (current_command<nold_commands && (old_commands[current_command].lastchar==';' || old_commands[current_command].lastchar=='/'))
    {
      for (x=strlen(command)-1;x>=0;x--)
      {
        if (strchr(" \t\n\r",command[x])==NULL)
        {
          if (command[x]==';' || command[x]=='/')
          {
            pos=x;
            lastchar=command[x];
            command[x]='\0';
            break;
          }
        }
      }
    }

    if ((current_command==nold_commands && command[0]!='\0') || (current_command<nold_commands && command[0]!='\0' && strcmp(old_commands[current_command].command,command)))
    {
      if (pos!=-1)
        command[pos]=lastchar;
      text->work_cmd.command=strdup(command);
      text->work_cmd.lastchar=0;
    }
    else
    {
      if (pos!=-1)
        command[pos]=lastchar;
    }
    current_command--;
    RecallCommand(text,&old_commands[current_command]);
  }
}

void NextStatement(TextWidget *text)
{
  if (tid)
    return;
  if (current_command<nold_commands-1)
  {
    current_command++;
    RecallCommand(text,&old_commands[current_command]);
  }
  else if (current_command==nold_commands-1)
  {
    current_command=nold_commands;
    if (text->work_cmd.command)
    {
      RecallCommand(text,&text->work_cmd);
      free(text->work_cmd.command);
      text->work_cmd.command=NULL;
    }
    else
      RecallCommand(text,NULL);
  }
}

void PrintErrorCharacter(int c)
{
	int x,firstchar=0;
  char *str;

  if (c>(int)strlen(command))
    return;
	str=strdup(command);
	for (x=c-1;x>=0;x--)
	{
		if (str[x]=='\n')
    {
			firstchar=x+1;
			break;
		}
	}
	for (x=c+1;x<(int)strlen(str);x++)
	{
		if (str[x]=='\n')
		{
			str[x]='\0';
		  break;
		}
	}
	c-=firstchar;
	WPrintf("\r%s\n",&str[firstchar]);
	WPrintf("%*s\n",c+1,"*");
  free(str);
}

void ShowPrompt(TextWidget *text)
{ 
 	WPrintf("%s ",STR_OPTION(PROMPT));
}

void InsertCommandLine(char *str)
{
	if (str==NULL)
	{
    command=StrdupCommand("\n");
	}
	else if (command==NULL)
	{
		command=StrdupCommand(str);
	}
	else
	{
    ReallocCommand(strlen(str)+strlen("\n")+sizeof(char));
/*  	command=realloc(command,strlen(command)+strlen(str)+strlen("\n")+sizeof(char));*/
	  strcat(command,"\n");
	  strcat(command,str);
	}
	nlines++;
}

void RedrawCommand(TextWidget *text,int pos,int nlines_before)
{
  int x,slen,lastx;
  char *str,new_prompt[100];

  if (nlines_before==1)
  {
  	sprintf(new_prompt,"\r%s ",STR_OPTION(PROMPT));
  }
  else
  {
  	sprintf(new_prompt,"\r%*d ",strlen(STR_OPTION(PROMPT)),nlines_before);
  }
	WPrintf(new_prompt);
  str=strdup(&command[pos]);
  slen=strlen(str);
  lastx=0;
  for (x=0;x<(int)slen;x++)
  {
    switch (str[x])
    {
      case '\n':
        str[x]='\0';
        WPrintfUP(&str[lastx]);
	      sprintf(new_prompt,"\n%*d ",strlen(STR_OPTION(PROMPT)),++nlines_before);
	      WPrintf("%s",new_prompt);
        lastx=x+1;
        break;
      default:
        break;
    }
  }
  if (lastx<slen)
    WPrintfUP(&str[lastx]);
  free(str);
}

void FreeTokens(void)
{
  int x;

  for (x=0;x<NTOKENS;x++)
  {
    if (old_tokens[x])
    {
      free(old_tokens[x]);
      old_tokens[x]=NULL;
    }
  }
  last_known_stmt=0;
}
  
void ClearCommand(TextWidget *text,int add,int sql_command)
{
	if (command!=NULL && command[0]!='\0')
	{
		if (text->scripts==0 && add && nlines<(int)INT_OPTION(MAXLINES) && (int)INT_OPTION(COMMANDHISTORY)>0)
		{
			Trim(command);
      if (INT_OPTION(HISTORY) && add)
        SaveToHistory(command);
      if (!nold_commands || strcmp(command,old_commands[nold_commands-1].command))
			{
				if (nold_commands==0)
				{
					old_commands=(CommandHistory*)malloc(sizeof(CommandHistory));
				}
				else
				{
          if (nold_commands<(int)INT_OPTION(COMMANDHISTORY))
          {
					  old_commands=(CommandHistory*)realloc(old_commands,(nold_commands+1)*sizeof(CommandHistory));
          }
          else
          {
            if ((int)INT_OPTION(COMMANDHISTORY)>1)
            {
              memmove(old_commands,&old_commands[1],(INT_OPTION(COMMANDHISTORY)-1)*sizeof(CommandHistory));
              if (nold_commands<(int)INT_OPTION(COMMANDHISTORY))
              {
                nold_commands=INT_OPTION(COMMANDHISTORY);
					      old_commands=(CommandHistory*)realloc(old_commands,(nold_commands+1)*sizeof(CommandHistory));
              }
            }
            nold_commands--;
          }
				}
				old_commands[nold_commands].command=strdup(command);
        old_commands[nold_commands].lastchar=lastchar;
        old_commands[nold_commands].sqlcommand=sql_command;
				nold_commands++;
        current_command=nold_commands;
			}
      else
        current_command=nold_commands;
		}
		FreeCommand();
	}
  command=StrdupCommand("");
  linepos=0;
  nlines=1;
  if (text->termout)
  {
    HANDLE tmp=text->spoolfile;

    text->spoolfile=NULL;
  	ShowPrompt(text);
    text->spoolfile=tmp;
  }
  if (add && text->work_cmd.command)
  {
    free(text->work_cmd.command);
    text->work_cmd.command=NULL;
  }
  FreeTokens();
}

int GetPreviousSQLCommand(void)
{
  int x;

  for (x=nold_commands-1;x>=0;x--)
  {
    if (old_commands[x].sqlcommand)
      return x;
  }
  return -1;
}


void EditCommand(TextWidget *text,char *filename)
{
  HANDLE f;
	DWORD bytes_written;
	PROCESS_INFORMATION p;
  STARTUPINFO si;
	int stopped=0,size,x,lastx=0,slen,ncommand;
	DWORD exit_code;
  char *str,command[MAX_PATH],new_filename[MAX_PATH];
	MSG  msg;

  ClearCommand(text,0,0);
  ncommand=GetPreviousSQLCommand();
  if (ncommand==-1)
	{
		WPrintf("No previous command\n");
		ShowPrompt(text);
		return;
	}
  if (filename==NULL)
  {
    f=CreateFile(STR_OPTION(EDIT_FILENAME),
               GENERIC_READ|GENERIC_WRITE,
               FILE_SHARE_READ,
               NULL,
               CREATE_ALWAYS,
               0,
               NULL);
             
    if (f==INVALID_HANDLE_VALUE)
    {
      DisplayLastError(NULL);
      WPrintf("Failed to create %s\n",STR_OPTION(EDIT_FILENAME));
		  ShowPrompt(text);
      return;
    }
    str=strdup(old_commands[ncommand].command);
    slen=strlen(str);
    for (x=0;x<(int)slen;x++)
    {
      switch (str[x])
      {
        case '\n':
          str[x]='\0';
          WriteFile(f,&str[lastx],strlen(&str[lastx]),&bytes_written,NULL);
          WriteFile(f,"\r\n",strlen("\r\n"),&bytes_written,NULL);
          lastx=x+1;
          break;
        default:
          break;
      }
    }
    if (lastx<slen)
      WriteFile(f,&str[lastx],strlen(&str[lastx]),&bytes_written,NULL);
    CloseHandle(f);
    free(str);
  }
  else
  {
    strcpy(new_filename,filename);
    f=CreateFile(new_filename,
               GENERIC_READ,
               FILE_SHARE_READ,
               NULL,
               OPEN_EXISTING,
               0,
               NULL);
             
    if (f==INVALID_HANDLE_VALUE)
    {
      if (strchr(new_filename,'.')==NULL)
        strcat(new_filename,".sql");
    }
    else
      CloseHandle(f);

  }
	si.cb=sizeof(STARTUPINFO);
	si.lpReserved=NULL;
	si.lpDesktop=NULL;
	si.lpTitle=NULL;
	si.dwFlags=0;
	si.cbReserved2=0;
	si.lpReserved2=NULL;
  sprintf(command,"%s %s",STR_OPTION(EDITOR),(filename==NULL)?(STR_OPTION(EDIT_FILENAME)):new_filename);
  if (CreateProcess(NULL,command,NULL,NULL,FALSE,0,NULL,STR_OPTION(DIRECTORY),&si,&p)==0)
  {
    DisplayLastError(NULL);
    WPrintf("Failed to spawn external editor\n");
		ShowPrompt(text);
    return;
  }
  tid=1;
	while (!stopped)
	{
	  while(PeekMessage( &msg, NULL, 0x00, 0x00, PM_REMOVE))
	  {
		  TranslateMessage(&msg);
		  DispatchMessage(&msg);
	  }
		GetExitCodeProcess(p.hProcess,&exit_code);
		if (exit_code!=STILL_ACTIVE)
			stopped=1;
	  Sleep(100);
	}
  tid=0;
	f=CreateFile((filename==NULL)?(STR_OPTION(EDIT_FILENAME)):new_filename,
             GENERIC_READ,
             FILE_SHARE_READ,
             NULL,
             OPEN_EXISTING,
             0,
             NULL);
             
  if (f==INVALID_HANDLE_VALUE)
  {
    DisplayLastError(NULL);
    WPrintf("Failed to open %s\n",(filename==NULL)?(STR_OPTION(EDIT_FILENAME)):new_filename);
		ShowPrompt(text);
    return;
  }
	size=GetFileSize(f,NULL);
	str=(char*)malloc(size+1);
  ReadFile(f,str,size,&bytes_written,NULL);
	CloseHandle(f);
  text->no_process=1;
	str[size]='\0';
  slen=strlen(str);
  if (slen && (str[slen-1]==';' || str[slen-1]=='/'))
    str[slen-1]='\0';
  slen=strlen(str);
  lastx=0;
  for (x=0;x<(int)slen;x++)
  {
    switch (str[x])
    {
      case '\r':
        str[x]='\0';
        if (strchr(&str[lastx],'\t'))
          InsertLineBuffer(text,"%s",&str[lastx]);
        else
          InsertUnparsedLineBuffer(text,&str[lastx]);
        InsertNewLineBuffer(text);
        lastx=x+2;
        break;
      default:
        break;
    }
  }
  if (lastx<slen)
  {
    if (strchr(&str[lastx],'\t'))
      InsertLineBuffer(text,"%s",&str[lastx]);
    else
      InsertUnparsedLineBuffer(text,&str[lastx]);
    InsertNewLineBuffer(text);
  }
  WPrintf("\r");
  text->no_process=0;
  ClearCommand(text,1,1);
}

int ExecuteCommand(ExecuteStruct *s)
{
	ClearList(hwndListView);
  CloseTMPFile();
	FreeCols(&cols,&ncols);
  if (s->INT_OPTION(AUTOTRACE))
    GetStatistics(&s->text->pre_stats,&s->text->npre_stats);
  if (s->data[0]=='#')
    DeleteCharacterFromString(s->data,0);
  StartTimer(s->text);
	nrows=DoCommand(s->data,s->INT_OPTION(FEEDBACK),s->executestr);
	if (nrows>=0)
	{
    if (s->INT_OPTION(FEEDBACK))
    {
      if (s->action[0]!='\0')
        WPrintf("\r%s %s\n",s->action,StopTimer(s->text));
      else
  	    WPrintf("\rStatement processed %s\n",StopTimer(s->text));
    }
    if (s->INT_OPTION(AUTOTRACE))
      ShowStatistics(s->text);
    if (s->INT_OPTION(SERVEROUTPUT))
      PrintServerOutput(s->text);
    if (s->text->scripts==0)
  	  tid=0;
    ClearCommand(s->text,1,1);
	}
	else
  {
    if (s->INT_OPTION(SERVEROUTPUT))
      PrintServerOutput(s->text);
    if (s->text->scripts==0)
  	  tid=0;
    ClearCommand(s->text,1,1);
  }
  free(s->data);
  free(s);
	return 1;
}

int ExecuteDML(ExecuteStruct *s)
{
	ClearList(hwndListView);
  CloseTMPFile();
	FreeCols(&cols,&ncols);
  if (s->INT_OPTION(AUTOTRACE))
    GetStatistics(&s->text->pre_stats,&s->text->npre_stats);
  if (s->data[0]=='#')
    DeleteCharacterFromString(s->data,0);
  StartTimer(s->text);
	nrows=DoCommand(s->data,s->INT_OPTION(FEEDBACK),NULL);
	if (nrows>=0)
	{
    if (s->INT_OPTION(FEEDBACK))
    {
      WPrintf("\r%d row%s %s %s\n",nrows,nrows==1?"":"s",s->action,StopTimer(s->text));
    }
    if (s->INT_OPTION(AUTOTRACE))
      ShowStatistics(s->text);
    if (s->INT_OPTION(SERVEROUTPUT))
      PrintServerOutput(s->text);
    if (s->text->scripts==0)
  	  tid=0;
    ClearCommand(s->text,1,1);
	}
	else
  {
    if (s->INT_OPTION(SERVEROUTPUT))
      PrintServerOutput(s->text);
    if (s->text->scripts==0)
  	  tid=0;
    ClearCommand(s->text,1,1);
  }
  free(s->data);
  free(s);
	return 1;
}

int ExecuteQuery(ExecuteStruct *s)
{
	ClearList(hwndListView);
	FreeCols(&cols,&ncols);
  if (s->INT_OPTION(AUTOTRACE))
    GetStatistics(&s->text->pre_stats,&s->text->npre_stats);
  StartTimer(s->text);
  if (s->data[0]=='#')
    DeleteCharacterFromString(s->data,0);
  if (!PrepareSource(s->text,s->data,&cols,&ncols,s->action))
  {
    if (s->action[0]!='\0' && s->INT_OPTION(FEEDBACK))
    {
      HANDLE tmp=s->text->spoolfile;

      s->text->spoolfile=NULL;
  	  WPrintfOnly("\rFetching data...");
      s->text->spoolfile=tmp;
    }
		if (!CreateTMPFile(s->text))
			nrows=FetchData(s->text,cols,ncols,s->action);
		else
			nrows=0;
    s->text->noredraw=0;
		if (nrows>=0 && s->action[0]!='\0')
    {
      if (s->INT_OPTION(FEEDBACK))
      {
  		  WPrintf(s->action,nrows,nrows==1?"":"s");
        WPrintf(" %s\n",StopTimer(s->text));
      }
    }
    if (s->INT_OPTION(LISTVIEW))
      InitListView(hwndListView);
    if (s->INT_OPTION(AUTOTRACE))
      ShowStatistics(s->text);
    if (s->INT_OPTION(SERVEROUTPUT))
      PrintServerOutput(s->text);
    if (s->text->scripts==0)
  	  tid=0;
  	ClearCommand(s->text,1,s->action[0]!='\0');
  }
	else
  {
    if (s->INT_OPTION(SERVEROUTPUT))
      PrintServerOutput(s->text);
    if (s->text->scripts==0)
  	  tid=0;
    ClearCommand(s->text,1,s->action[0]!='\0');
  }
  FinishSource(cols,ncols);
  free(s->data);
  free(s);
	return 1;
}

void CreateNewToken(char **tokens,int *ntokens,char *str,int lastpos,int pos)
{
  str[pos]='\0';
  tokens[*ntokens]=&str[lastpos];
  (*ntokens)++;
}

int Tokenize(char *str,char **tokens)
{
  int lastpos=0,x,ntokens=0,len;
  char quote=0;

  len=strlen(str);
  for (x=0;x<=len && ntokens<NTOKENS;x++)
  {
    if (!quote && strchr(" \t\n\r()",str[x])!=NULL)
    {
      if (lastpos!=x)
        CreateNewToken(tokens,&ntokens,str,lastpos,x);
      lastpos=x+1;
    }
    else if ((str[x]=='"' || str[x]=='\'') && (!quote || str[x]==quote))
    {
      if (!quote)
      {
        quote=str[x];
        if (x && str[x-1]=='@') /* Special case for @"c:\program files\script.sql" */
          CreateNewToken(tokens,&ntokens,str,lastpos,x);
      }
      else
      {
        CreateNewToken(tokens,&ntokens,str,lastpos,x);
        quote=0;
      }
      lastpos=x+1;
    }
  }
  return ntokens;
}

int InsideComment(char *command)
{
  int x;
  char *str=command;

  for (x=strlen(command)-1;x>=0;x--)
  {
    if (command[x]=='\n')
      break;
    if (!strncmp(&command[x],"--",2))
      return 1;
  }
  while (str)
  {
    str=strstr(str,"/*");
    if (str==NULL)
      return 0;
    str=strstr(str,"*/");
    if (str==NULL)
      return 1;
  }
  return 0;
}

void RememberObject(char *type,char *name)
{
  if (last_object_type!=NULL)
    free(last_object_type);
  last_object_type=strdup(type);
  if (last_object_name!=NULL)
    free(last_object_name);
  last_object_name=strdup(name);
}

void RememberTokens(char **tokens)
{
  int x;

  for (x=0;x<NTOKENS;x++)
  {
    old_tokens[x]=strdup(tokens[x]);
  }
}

int ProcessCommand(TextWidget *text)
{
	static int x,ntokens;
  int lastpos=0;
  char *new_command=NULL;
  char *(tokens[NTOKENS])={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  static ExecuteStruct *es;
  int return_value=0;

  if (command==NULL)
		return 0;

  lastchar='\n';
	for (x=strlen(command)-1;x>=0;x--)
	{
		if (command[x]!=' ' && command[x]!='\n' && command[x]!='\t')
		{
			lastchar=command[x];
			lastpos=x;
			break;
		}
	}
  if (old_tokens[0])
  {
    ntokens=NTOKENS;
    memcpy(&tokens,&old_tokens,sizeof(old_tokens));
  }
  else
  {
    new_command=strdup(command);
    if (new_command[0]=='#')
      DeleteCharacterFromString(new_command,0);
    if (lastchar==';')
      new_command[lastpos]='\0';
    ntokens=Tokenize(new_command,(char**)tokens);
    if (ntokens==NTOKENS)
      RememberTokens(tokens);
    if (ntokens==0)
    {
		  ClearCommand(text,0,0);
      if (new_command)
        free(new_command);
		  return 2;
	  }
  }
	if (tokens[0][strlen(tokens[0])-1]==';')
		tokens[0][strlen(tokens[0])-1]='\0';

	/* Start of PL/SQL block */
  if (last_known_stmt==1 ||
      !_stricmp(tokens[0],"begin") || 
      !_stricmp(tokens[0],"declare") ||
      (ntokens>=4 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"package")) ||
      (ntokens>=2 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"procedure")) ||
      (ntokens>=4 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"procedure")) ||
      (ntokens>=2 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"function")) ||
      (ntokens>=4 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"function")) ||
      (ntokens>=2 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"package")) ||
      (ntokens>=2 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"type")) ||
      (ntokens>=4 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"type")) ||
      (ntokens>=2 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"trigger")) ||
      (ntokens>=4 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"trigger")))
	{
    if (lastchar=='/' && (lastpos==0 || command[lastpos-1]!='*') && !InsideComment(command))
		{
			command[lastpos]='\0';
      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      es->data=strdup(command);
      strcpy(es->action,"");
      strcpy(es->executestr,"");
      if (ntokens>=3 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"procedure"))
      {
        sprintf(es->action,"Procedure %s created",tokens[2]);
        sprintf(es->executestr,"Creating procedure %s",tokens[2]);
        RememberObject("PROCEDURE",tokens[2]);
      }
      else if (ntokens>=5 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"procedure"))
      {
        sprintf(es->action,"Procedure %s created",tokens[4]);
        sprintf(es->executestr,"Creating procedure %s",tokens[4]);
        RememberObject("PROCEDURE",tokens[4]);
      }
      else if (ntokens>=3 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"function"))
      {
        sprintf(es->action,"Function %s created",tokens[2]);
        sprintf(es->executestr,"Creating function %s",tokens[2]);
        RememberObject("FUNCTION",tokens[2]);
      }
      else if (ntokens>=5 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"function"))
      {
        sprintf(es->action,"Function %s created",tokens[4]);
        sprintf(es->executestr,"Creating function %s",tokens[4]);
        RememberObject("FUNCTION",tokens[4]);
      }
      else if (ntokens>=4 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"package") && !_stricmp(tokens[2],"body"))
      {
        sprintf(es->action,"Package body %s created",tokens[3]);
        sprintf(es->executestr,"Creating package body %s",tokens[3]);
        RememberObject("PACKAGE BODY",tokens[3]);
      }
      else if (ntokens>=6 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"package") && !_stricmp(tokens[4],"body"))
      {
        sprintf(es->action,"Package body %s created",tokens[5]);
        sprintf(es->executestr,"Creating package body %s",tokens[5]);
        RememberObject("PACKAGE BODY",tokens[5]);
      }
      else if (ntokens>=3 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"package"))
      {
        sprintf(es->action,"Package %s created",tokens[2]);
        sprintf(es->executestr,"Creating package %s",tokens[2]);
        RememberObject("PACKAGE",tokens[2]);
      }
      else if (ntokens>=5 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"package"))
      {
        sprintf(es->action,"Package %s created",tokens[4]);
        sprintf(es->executestr,"Creating package %s",tokens[4]);
        RememberObject("PACKAGE",tokens[4]);
      }
      else if (ntokens>=4 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"type") && !_stricmp(tokens[2],"body"))
      {
        sprintf(es->action,"Type body %s created",tokens[3]);
        sprintf(es->executestr,"Creating type body %s",tokens[3]);
        RememberObject("TYPE BODY",tokens[3]);
      }
      else if (ntokens>=6 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"type") && !_stricmp(tokens[4],"body"))
      {
        sprintf(es->action,"Type body %s created",tokens[5]);
        sprintf(es->executestr,"Creating type body %s",tokens[5]);
        RememberObject("TYPE BODY",tokens[5]);
      }
      else if (ntokens>=3 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"type"))
      {
        sprintf(es->action,"Type %s created",tokens[2]);
        sprintf(es->executestr,"Creating type %s",tokens[2]);
        RememberObject("TYPE",tokens[2]);
      }
      else if (ntokens>=5 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"type"))
      {
        sprintf(es->action,"Type %s created",tokens[4]);
        sprintf(es->executestr,"Creating type %s",tokens[4]);
        RememberObject("TYPE",tokens[4]);
      }
      else if (ntokens>=3 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"trigger"))
      {
        sprintf(es->action,"Trigger %s created",tokens[2]);
        sprintf(es->executestr,"Creating trigger %s",tokens[2]);
        RememberObject("TRIGGER",tokens[2]);
      }
      else if (ntokens>=5 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"trigger"))
      {
        sprintf(es->action,"Trigger %s created",tokens[4]);
        sprintf(es->executestr,"Creating trigger %s",tokens[4]);
        RememberObject("TRIGGER",tokens[4]);
      }
      else if (!_stricmp(tokens[0],"begin") || !_stricmp(tokens[0],"begin"))
      {
        strcpy(es->executestr,"Executing PL/SQL procedure");
        strcpy(es->action,"PL/SQL procedure successfully completed");
      }
      if (tid==0)
  			CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteCommand,es,0,&tid);
      else
        ExecuteCommand(es);
      allow_blank_lines=0;
      return_value=1;
		}
    else
    {
      allow_blank_lines=1;
      if (last_known_stmt==0 && ntokens==10)
        last_known_stmt=1;
    }
  }
  else if (!_stricmp(tokens[0],"exec") || !_stricmp(tokens[0],"execu") || !_stricmp(tokens[0],"execut") || !_stricmp(tokens[0],"execute"))
	{
    if (lastchar==';')
		{
			command[lastpos]='\0';
      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      es->data=malloc(strlen(command)+20);
      strcpy(es->data,"BEGIN ");
      strcat(es->data,&command[strlen(tokens[0])]+1);
      strcat(es->data,"; END;");
      strcpy(es->action,"PL/SQL procedure successfully completed");
      strcpy(es->executestr,"Executing PL/SQL procedure");
      if (tid==0)
  			CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteCommand,es,0,&tid);
      else
        ExecuteCommand(es);
      allow_blank_lines=0;
      return_value=1;
		}
  }
  else if ((ntokens>=2 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"view")) ||
           (ntokens>=4 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"view")) ||
           (ntokens>=3 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"force") && !_stricmp(tokens[2],"view")) ||
           (ntokens>=5 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"force") && !_stricmp(tokens[4],"view")) ||
           (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"view") && !_stricmp(tokens[3],"compile")) ||
           (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"procedure") && !_stricmp(tokens[3],"compile")) ||
           (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"function") && !_stricmp(tokens[3],"compile")) ||
           (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"package") && !_stricmp(tokens[3],"compile")) ||
           (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"type") && !_stricmp(tokens[3],"compile")) ||
           (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"trigger") && !_stricmp(tokens[3],"compile")) ||
           (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"trigger") && !_stricmp(tokens[3],"compile")))
	{
    if (lastchar=='/' || lastchar==';')
		{
			command[lastpos]='\0';
      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      es->data=strdup(command);
      strcpy(es->action,"");
      strcpy(es->executestr,"");
      if (ntokens>=3 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"view"))
      {
        sprintf(es->action,"View %s created",tokens[2]);
        sprintf(es->executestr,"Creating view %s",tokens[2]);
        RememberObject("VIEW",tokens[2]);
      }
      else if (ntokens>=5 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"view"))
      {
        sprintf(es->action,"View %s created",tokens[4]);
        sprintf(es->executestr,"Creating view %s",tokens[4]);
        RememberObject("VIEW",tokens[4]);
      }
      else if (ntokens>=4 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"force") && !_stricmp(tokens[2],"view"))
      {
        sprintf(es->action,"View %s created",tokens[3]);
        sprintf(es->executestr,"Creating view %s",tokens[3]);
        RememberObject("VIEW",tokens[3]);
      }
      else if (ntokens>=6 && !_stricmp(tokens[0],"create") && !_stricmp(tokens[1],"or") && !_stricmp(tokens[2],"replace") && !_stricmp(tokens[3],"force") && !_stricmp(tokens[4],"view"))
      {
        sprintf(es->action,"View %s created",tokens[5]);
        sprintf(es->executestr,"Creating view %s",tokens[5]);
        RememberObject("VIEW",tokens[5]);
      }
      else if (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"view") && !_stricmp(tokens[3],"compile"))
      {
        sprintf(es->action,"View %s altered",tokens[2]);
        sprintf(es->executestr,"Altering view %s",tokens[2]);
        RememberObject("VIEW",tokens[2]);
      }
      else if (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"procedure") && !_stricmp(tokens[3],"compile"))
      {
        sprintf(es->action,"Procedure %s altered",tokens[2]);
        sprintf(es->executestr,"Altering procedure %s",tokens[2]);
        RememberObject("PROCEDURE",tokens[2]);
      }
      else if (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"function") && !_stricmp(tokens[3],"compile"))
      {
        sprintf(es->action,"Function %s altered",tokens[2]);
        sprintf(es->executestr,"Altering function %s",tokens[2]);
        RememberObject("FUNCTION",tokens[2]);
      }
      else if (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"trigger") && !_stricmp(tokens[3],"compile"))
      {
        sprintf(es->action,"Trigger %s altered",tokens[2]);
        sprintf(es->executestr,"Altering trigger %s",tokens[2]);
        RememberObject("TRIGGER",tokens[2]);
      }
      else if (ntokens>=5 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"package") && !_stricmp(tokens[3],"compile") && !_stricmp(tokens[4],"body"))
      {
        sprintf(es->action,"Package body %s altered",tokens[2]);
        sprintf(es->executestr,"Altering package body %s",tokens[2]);
        RememberObject("PACKAGE BODY",tokens[2]);
      }
      else if (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"package") && !_stricmp(tokens[3],"compile"))
      {
        sprintf(es->action,"Package %s altered",tokens[2]);
        sprintf(es->executestr,"Altering package %s",tokens[2]);
        RememberObject("PACKAGE",tokens[2]);
      }
      else if (ntokens>=5 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"type") && !_stricmp(tokens[3],"compile") && !_stricmp(tokens[4],"body"))
      {
        sprintf(es->action,"Type body %s altered",tokens[2]);
        sprintf(es->executestr,"Altering type body %s",tokens[2]);
        RememberObject("TYPE BODY",tokens[2]);
      }
      else if (ntokens>=4 && !_stricmp(tokens[0],"alter") && !_stricmp(tokens[1],"type") && !_stricmp(tokens[3],"compile"))
      {
        sprintf(es->action,"Type %s altered",tokens[2]);
        sprintf(es->executestr,"Altering type %s",tokens[2]);
        RememberObject("TYPE",tokens[2]);
      }
      if (tid==0)
  			CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteCommand,es,0,&tid);
      else
        ExecuteCommand(es);
      return_value=1;
		}
  }
	else if (!_stricmp(tokens[0],"select") || !_stricmp(tokens[0],"with"))
	{
		if (lastchar==';' || (lastchar=='/' && command[lastpos-1]!='*'))
		{
			command[lastpos]='\0';
      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      es->data=strdup(command);
      strcpy(es->action,"\r%d row%s selected");
      strcpy(es->executestr,"");
      if (es->INT_OPTION(AUTOTRACE))
        ShowAutotrace(es);
      if (tid==0)
  			CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteQuery,es,0,&tid);
      else
        ExecuteQuery(es);
      return_value=2;
		}
	}
	else if (!_stricmp(tokens[0],"update"))
	{
		if (lastchar==';' || (lastchar=='/' && command[lastpos-1]!='*'))
		{
			command[lastpos]='\0';
      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      es->data=strdup(command);
      strcpy(es->action,"updated");
      strcpy(es->executestr,"");
      if (es->INT_OPTION(AUTOTRACE))
        ShowAutotrace(es);
      if (tid==0)
  			CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteDML,es,0,&tid);
      else
        ExecuteDML(es);
      return_value=1;
		}
	}
	else if (!_stricmp(tokens[0],"insert"))
	{
		if (lastchar==';' || (lastchar=='/' && command[lastpos-1]!='*'))
		{
			command[lastpos]='\0';
      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      es->data=strdup(command);
      strcpy(es->action,"inserted");
      strcpy(es->executestr,"");
      if (es->INT_OPTION(AUTOTRACE))
        ShowAutotrace(es);
      if (tid==0)
  			CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteDML,es,0,&tid);
      else
        ExecuteDML(es);
      return_value=1;
		}
	}
	else if (!_stricmp(tokens[0],"delete"))
	{
		if (lastchar==';' || (lastchar=='/' && command[lastpos-1]!='*'))
		{
			command[lastpos]='\0';
      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      es->data=strdup(command);
      strcpy(es->action,"deleted");
      strcpy(es->executestr,"");
      if (es->INT_OPTION(AUTOTRACE))
        ShowAutotrace(es);
      if (tid==0)
  			CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteDML,es,0,&tid);
      else
        ExecuteDML(es);
      return_value=1;
    }
	}
	else if (!_stricmp(tokens[0],"exit"))
	{
    if (tid)
      PostMessage(g_hWnd,WM_QUIT,(WPARAM)0,(LPARAM)0);
    else
  		PostQuitMessage(0);
    ClearCommand(text,0,0);
    return_value=2;
  }
	else if (!_stricmp(tokens[0],"commit"))
	{
		if (lastchar==';' || lastchar=='/')
		{
			command[lastpos]='\0';
		  Commit();
		  ClearCommand(text,1,0);
      return_value=2;
    }
  }
	else if (!_stricmp(tokens[0],"rollback") && ntokens==1)
	{
		if (lastchar==';' || lastchar=='/')
		{
			command[lastpos]='\0';
		  Rollback();
		  ClearCommand(text,1,0);
      return_value=2;
    }
  }
	else if (!_stricmp(tokens[0],"dir") || !_stricmp(tokens[0],"ls"))
	{
    es=malloc(sizeof(ExecuteStruct));
    es->text=text;
    es->data=strdup("select object_name,object_type,status from user_objects order by 2,1");
    strcpy(es->action,"");
    strcpy(es->executestr,"");
    if (tid==0)
  		CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteQuery,es,0,&tid);
    else
      ExecuteQuery(es);
    return_value=1;
  }
	else if (!_stricmp(tokens[0],"r"))
	{
    int ncommand;

		FreeCommand();
		command=StrdupCommand("");
    linepos=0;
    ncommand=GetPreviousSQLCommand();
		if (ncommand!=-1 && text->no_process==0)
    {
      ClearCommand(text,0,0);
      text->no_process=1;
      InsertLineBuffer(text,old_commands[ncommand].command);
      InsertNewLineBuffer(text);
			InsertCommandLine("/");
      if (text->termout)
        WPrintf("\r");
      ProcessCommand(text);
      text->no_process=0;
      return_value=2;
    }
		else
		{
			WPrintf("No previous command\n");
			ClearCommand(text,0,0);
      return_value=2;
		}
  }
	else if (!_stricmp(tokens[0],"/"))
	{
    int ncommand;

		FreeCommand();
    linepos=0;
    ncommand=GetPreviousSQLCommand();
		if (ncommand!=-1 && text->no_process==0)
    {
      text->no_process=1;
      command=StrdupCommand(old_commands[ncommand].command);
			InsertCommandLine("/");
      ProcessCommand(text);
      text->no_process=0;
      return_value=2;
    }
		else
		{
  		command=StrdupCommand("");
			WPrintf("No previous command\n");
			ClearCommand(text,0,0);
      return_value=2;
		}
  }
	else if (!_stricmp(tokens[0],"ed") || !_stricmp(tokens[0],"edi") || !_stricmp(tokens[0],"edit"))
	{
    if (ntokens==1)
  		EditCommand(text,NULL);
    else if (ntokens==2)
      EditCommand(text,tokens[1]);
    else
    {
      WPrintf("Usage: edit [filename]\n");
      ClearCommand(text,1,0);
    }
    return_value=2;
	}
	else if (!_stricmp(tokens[0],"desc") || !_stricmp(tokens[0],"descr") || !_stricmp(tokens[0],"descri") || !_stricmp(tokens[0],"describ") || !_stricmp(tokens[0],"describe"))
	{
		if (lastchar==';' || lastchar=='/')
		{
			command[lastpos]='\0';
    }
    Describe(text,command);
    return_value=2;
	}
  else if (!_stricmp(tokens[0],"show") && ntokens>=2 &&
          (!_stricmp(tokens[1],"err") || !_stricmp(tokens[1],"erro") || !_stricmp(tokens[1],"error") || !_stricmp(tokens[1],"errors")))
  {
		if (lastchar==';' || lastchar=='/')
		{
			command[lastpos]='\0';
    }
    if (ntokens==2)
    {
      if (last_object_type==NULL)
      {
        WPrintf("No known object compiled\n");
        ClearCommand(text,1,0);
      }
      else
        ShowError(text,last_object_type,last_object_name);
    }
    else if (ntokens==4)
    {
      ShowError(text,tokens[2],tokens[3]);
    }
    else if (ntokens==5)
    {
      char *tmpstr;

      tmpstr=(char*)malloc(strlen(tokens[2])+strlen(tokens[3])+2);
      if (tmpstr==NULL)
      {
        WPrintf("Out of memory\n");
        ClearCommand(text,1,0);
      }
      else
      {
        sprintf(tmpstr,"%s %s",tokens[2],tokens[3]);
        ShowError(text,tmpstr,tokens[4]);
        free(tmpstr);
      }
    }
    else
    {
      WPrintf("Usage: show err[or] [object_type object_name]\n");
      ClearCommand(text,1,0);
    }
    return_value=2;
  }
  else if (!strncmp(tokens[0],"@",1))
  {
    if (text->scripts>=(int)INT_OPTION(MAX_SCRIPTS))
    {
      WPrintf("Maximum number of nested scripts (%d) exceeded\n",INT_OPTION(MAX_SCRIPTS));
    }
    else
    {
      if (lastchar=='/' || lastchar==';')
        command[lastpos]='\0';
      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      es->data=strdup(command);
      strcpy(es->action,"");
      strcpy(es->executestr,"");
      if (tid==0)
  		  CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)RunScript,es,0,&tid);
      else
        RunScript(es);
    }
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"spool"))
  {
    if (ntokens!=2)
    {
      WPrintf("Usage: spool file|off\n");
    }
    else
    {
      if (tokens[1][strlen(tokens[1]-1)]==';')
        tokens[1][strlen(tokens[1]-1)]='\0';
      if (!_stricmp(tokens[1],"off"))
      {
        if (text->spoolfile==NULL)
          WPrintf("Not currently spooling\n");
        else
        {
          CloseHandle(text->spoolfile);
          text->spoolfile=NULL;
        }
      }
      else
      {
        if (text->spoolfile)
          CloseHandle(text->spoolfile);
        text->spoolfile=CreateFile(tokens[1],
                                   GENERIC_WRITE,
                                   FILE_SHARE_READ,
                                   NULL,
                                   CREATE_ALWAYS,
                                   FILE_ATTRIBUTE_NORMAL,
                                   NULL);
        if (text->spoolfile==INVALID_HANDLE_VALUE)
        {
          text->spoolfile=NULL;
          DisplayLastError(NULL);
        }
      }
    }
		ClearCommand(text,1,0);
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"set") && 
           (ntokens==1 ||
           (_stricmp(tokens[1],"constraint") &&
            _stricmp(tokens[1],"constraints") &&
            _stricmp(tokens[1],"role") &&
            _stricmp(tokens[1],"transaction"))))
  {
    if (ntokens==1)
      PrintOptions(text);
    else if (ntokens<3)
    {
      WPrintf("Usage: set parameter value [parameter value ...]\n");
    }
    else
    {
      SetCommandLineOption(text,&command[4]);
    }
		ClearCommand(text,1,0);
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"define"))
  {
    if (ntokens>=2)
      DefineVariable(&command[7]);
    else
      WPrintf("Usage: define variable[=value]\n");
		ClearCommand(text,1,0);
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"undef") || !_stricmp(tokens[0],"undefi") || !_stricmp(tokens[0],"undefin") || !_stricmp(tokens[0],"undefine"))
  {
    if (ntokens<2)
    {
      WPrintf("Usage: undefine variable\n");
    }
    else
    {
      if (!HTDeleteRecord(vht,&tokens[1]))
        WPrintf("Variable %s does not exist\n",tokens[1]);
    }
		ClearCommand(text,1,0);
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"acc") || !_stricmp(tokens[0],"acce") || !_stricmp(tokens[0],"accep") || !_stricmp(tokens[0],"accept"))
  {
    AcceptVariable(text,tokens,ntokens);
		ClearCommand(text,1,0);
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"conn") || !_stricmp(tokens[0],"conne") || !_stricmp(tokens[0],"connec") || !_stricmp(tokens[0],"connect"))
  {
    if (ntokens==1 || ConnectWithConnectString(&command[strlen(tokens[0])])==1)
    {
      WPrintf("Usage: connect connectstring [as normal|sysoper|sysdba]\n");
    }
    if (ora_connected)
    {
      GetDefaultOptions(g_text);
      SetPWD(g_text);
    }
		ClearCommand(text,1,0);
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"disc") || !_stricmp(tokens[0],"disco") || !_stricmp(tokens[0],"discon") || !_stricmp(tokens[0],"disconn") || !_stricmp(tokens[0],"disconne") || !_stricmp(tokens[0],"disconnec") ||  !_stricmp(tokens[0],"disconnect"))
  {
    DisconnectFromOracle();
    if (text->termout)
      WPrintf("Disconnected\n");
		ClearCommand(text,1,0);
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"prompt"))
  {
    if (ntokens>1)
      WPrintf("%s\n",&command[7]);
		ClearCommand(text,1,0);
    return_value=2;
  }
  else if (!_strnicmp(tokens[0],"--",2) || !_stricmp(tokens[0],"rem"))
  {
		ClearCommand(text,0,0);
    return_value=2;
  }
  else if (!_strnicmp(tokens[0],"/*",2))
  {
    char * tmpstr;

    tmpstr=strdup(command);
    Trim(tmpstr);
    if (!strcmp(&tmpstr[strlen(tmpstr)-2],"*/"))
    {
      allow_blank_lines=0;
      ClearCommand(text,0,0);
      free(tmpstr);
      return_value=2;
    }
    else
    {
      allow_blank_lines=1;
      free(tmpstr);
      return_value=0;
    }
  }
  else if (!_stricmp(tokens[0],"tune"))
  {
    int ncommand;

    ncommand=GetPreviousSQLCommand();
    if (ncommand!=-1)
    {
      if (text->analyze==NULL)
        text->analyze=ShowSqlAnalyzer(NULL);
      SetAnalyzerText(text,old_commands[ncommand].command);
      ExplainPlan(text->analyze,old_commands[ncommand].command);
  		ClearCommand(text,1,0);
    }
    else
    {
      WPrintf("No previous command\n");
      ClearCommand(text,1,0);
    }
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"col") || !_stricmp(tokens[0],"colu") || !_stricmp(tokens[0],"colum") || !_stricmp(tokens[0],"column"))
  { /* Just ignore this one for the moment */
		ClearCommand(text,1,0);
    return_value=2;
  }
  else if (!_stricmp(tokens[0],"show") && ntokens>1 &&
          (!_stricmp(tokens[1],"par") ||
           !_stricmp(tokens[1],"para") ||
           !_stricmp(tokens[1],"param") ||
           !_stricmp(tokens[1],"parame") ||
           !_stricmp(tokens[1],"paramet") ||
           !_stricmp(tokens[1],"paramete") ||
           !_stricmp(tokens[1],"parameter")))
  {
    if (ntokens==3)
    {
      char *tmpstr;

      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      tmpstr=(char*)malloc(60+strlen(tokens[2]));
      if (tmpstr==NULL)
      {
        WPrintf("Out of memory\n");
      }
      else
      {
        sprintf(tmpstr,"select * from v$parameter where name like lower('%%%s%%')",tokens[2]);
        es->data=strdup(tmpstr);
        strcpy(es->action,"");
        strcpy(es->executestr,"");
        if (tid==0)
  		    CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteQuery,es,0,&tid);
        else
          ExecuteQuery(es);
        free(tmpstr);
        return_value=1;
      }
    }
    else
    {
      WPrintf("Usage: show par[ameter] name\n");
      ClearCommand(text,1,0);
      return_value=2;
    }
  }
	else
	{
		if (lastchar==';' || lastchar=='/')
		{
			command[lastpos]='\0';
      es=malloc(sizeof(ExecuteStruct));
      es->text=text;
      es->data=strdup(command);
      strcpy(es->executestr,"");
      if (!_stricmp(tokens[0],"create") && ntokens>2 &&
               (!_stricmp(tokens[1],"cluster") ||
                !_stricmp(tokens[1],"dimension") ||
                !_stricmp(tokens[1],"directory") ||
                !_stricmp(tokens[1],"index") ||
                !_stricmp(tokens[1],"indextype") ||
                !_stricmp(tokens[1],"library") ||
                !_stricmp(tokens[1],"operator") ||
                !_stricmp(tokens[1],"profile") ||
                !_stricmp(tokens[1],"role") ||
                !_stricmp(tokens[1],"sequence") ||
                !_stricmp(tokens[1],"synonym") ||
                !_stricmp(tokens[1],"table") ||
                !_stricmp(tokens[1],"tablespace") ||
                !_stricmp(tokens[1],"user")
               ))
      {
        ToLower(tokens[1]);
        ToUpper(tokens[2]);
        sprintf(es->action,"%s %s created",tokens[1],tokens[2]);
        sprintf(es->executestr,"Creating %s %s",tokens[1],tokens[2]);
      }
      else if (!_stricmp(tokens[0],"create") && ntokens>3 &&
               !_stricmp(tokens[1],"public") &&
               !_stricmp(tokens[2],"synonym"))
      {
        ToLower(tokens[1]);
        ToUpper(tokens[3]);
        sprintf(es->action,"public synonym %s created",tokens[3]);
        sprintf(es->executestr,"Creating public synonym %s",tokens[3]);
      }
      else if (!_stricmp(tokens[0],"drop") && ntokens>3 &&
               !_stricmp(tokens[1],"public") &&
               !_stricmp(tokens[2],"synonym"))
      {
        ToLower(tokens[1]);
        ToUpper(tokens[3]);
        sprintf(es->action,"public synonym %s dropped",tokens[3]);
        sprintf(es->executestr,"Dropping public synonym %s",tokens[3]);
      }
      else if (!_stricmp(tokens[0],"drop") && ntokens>2 &&
               (!_stricmp(tokens[1],"cluster") ||
                !_stricmp(tokens[1],"dimension") ||
                !_stricmp(tokens[1],"directory") ||
                !_stricmp(tokens[1],"function") ||
                !_stricmp(tokens[1],"index") ||
                !_stricmp(tokens[1],"indextype") ||
                !_stricmp(tokens[1],"library") ||
                !_stricmp(tokens[1],"operator") ||
                !_stricmp(tokens[1],"package") ||
                !_stricmp(tokens[1],"procedure") ||
                !_stricmp(tokens[1],"profile") ||
                !_stricmp(tokens[1],"role") ||
                !_stricmp(tokens[1],"sequence") ||
                !_stricmp(tokens[1],"synonym") ||
                !_stricmp(tokens[1],"table") ||
                !_stricmp(tokens[1],"tablespace") ||
                !_stricmp(tokens[1],"trigger") ||
                !_stricmp(tokens[1],"user") ||
                !_stricmp(tokens[1],"view")
               ))
      {
        ToLower(tokens[1]);
        ToUpper(tokens[2]);
        sprintf(es->action,"%s %s dropped",tokens[1],tokens[2]);
        sprintf(es->executestr,"Dropping %s %s",tokens[1],tokens[2]);
      }
      else if (!_stricmp(tokens[0],"alter") && ntokens>2 &&
               (!_stricmp(tokens[1],"cluster") ||
                !_stricmp(tokens[1],"dimension") ||
                !_stricmp(tokens[1],"index") ||
                !_stricmp(tokens[1],"indextype") ||
                !_stricmp(tokens[1],"profile") ||
                !_stricmp(tokens[1],"role") ||
                !_stricmp(tokens[1],"sequence") ||
                !_stricmp(tokens[1],"table") ||
                !_stricmp(tokens[1],"type") ||
                !_stricmp(tokens[1],"user")
               ))
      {
        ToLower(tokens[1]);
        ToUpper(tokens[2]);
        sprintf(es->action,"%s %s altered",tokens[1],tokens[2]);
        sprintf(es->executestr,"Altering %s %s",tokens[1],tokens[2]);
      }
      else if (!_stricmp(tokens[0],"analyze") && ntokens>2 &&
               (!_stricmp(tokens[1],"cluster") ||
                !_stricmp(tokens[1],"index") ||
                !_stricmp(tokens[1],"table")
               ))
      {
        ToLower(tokens[1]);
        ToUpper(tokens[2]);
        sprintf(es->action,"%s %s analyzed",tokens[1],tokens[2]);
        sprintf(es->executestr,"Analyzing %s %s",tokens[1],tokens[2]);
      }
      else if (!_stricmp(tokens[0],"comment") && !_stricmp(tokens[1],"on") && ntokens>3 &&
               (!_stricmp(tokens[2],"table") ||
                !_stricmp(tokens[2],"column") ||
                !_stricmp(tokens[2],"operator") ||
                !_stricmp(tokens[2],"indextype")
               ))
      {
        ToLower(tokens[2]);
        ToUpper(tokens[3]);
        sprintf(es->action,"comment on %s %s created",tokens[2],tokens[3]);
        sprintf(es->executestr,"Creating comment on %s %s",tokens[2],tokens[3]);
      }
      else if (!_stricmp(tokens[0],"explain") && !_stricmp(tokens[1],"on"))
      {
        sprintf(es->action,"plan explained");
        sprintf(es->executestr,"Explaining plan");
      }
      else if (!_stricmp(tokens[0],"grant"))
      {
        sprintf(es->action,"Grant succeeded");
        sprintf(es->executestr,"Granting");
      }
      else if (!_stricmp(tokens[0],"rollback"))
      {
        sprintf(es->action,"Rollback complete");
        sprintf(es->executestr,"Rolling back");
      }
      else if (!_stricmp(tokens[0],"revoke"))
      {
        sprintf(es->action,"Revoke succeeded");
        sprintf(es->executestr,"Revoking");
      }
      else if (!_stricmp(tokens[0],"lock") && !_stricmp(tokens[1],"table") && ntokens>3)
      {
        ToUpper(tokens[2]);
        sprintf(es->action,"table %s is locked",tokens[2]);
        sprintf(es->executestr,"Locking table %s",tokens[2]);
      }
      else if (!_stricmp(tokens[0],"rename") && !_stricmp(tokens[2],"to") && ntokens==4)
      {
        ToUpper(tokens[1]);
        ToUpper(tokens[3]);
        sprintf(es->action,"Object %s is renamed to %s",tokens[1],tokens[3]);
        sprintf(es->executestr,"Renaming object %s to %s",tokens[1],tokens[3]);
      }
      else if (!_stricmp(tokens[0],"savepoint") && ntokens==2)
      {
        ToUpper(tokens[1]);
        sprintf(es->action,"savepoint %s set",tokens[1]);
        sprintf(es->executestr,"Setting savepoint %s",tokens[1]);
      }
      else if (!_stricmp(tokens[0],"truncate") && ntokens>2 &&
               (!_stricmp(tokens[1],"cluster") ||
                !_stricmp(tokens[1],"table")
               ))
      {
        ToLower(tokens[1]);
        ToUpper(tokens[2]);
        sprintf(es->action,"%s %s truncated",tokens[1],tokens[2]);
        sprintf(es->executestr,"Truncating %s %s",tokens[1],tokens[2]);
      }
      else
      {
        strcpy(es->action,"");
        strcpy(es->executestr,"");
      }
      if (tid==0)
  			CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)ExecuteCommand,es,0,&tid);
      else
        ExecuteCommand(es);
      return_value=1;
		}
	}
  if (new_command)
    free(new_command);
  return return_value;
}