#include "jbsql.h"

#define ITEM_COUNT 1000

int  last_ncols=0;
char *buffer=NULL;

HWND CreateListView(HINSTANCE hInstance, HWND hwndParent)
{
	DWORD       dwStyle;
	HWND        hwndListView;
	BOOL        bSuccess = TRUE;

	dwStyle =   WS_TABSTOP | 
							WS_CHILD| 
							WS_VISIBLE |
							LVS_REPORT | 
							LVS_OWNERDATA|
							LVS_NOSORTHEADER;
            
	hwndListView = CreateWindowEx(   WS_EX_CLIENTEDGE,          // ex style
																	 WC_LISTVIEW,               // class name - defined in commctrl.h
																	 "",                        // dummy text
																	 dwStyle,                   // style
																	 0,                         // x position
																	 0,                         // y position
																	 0,                         // width
																	 0,                         // height
																	 hwndParent,                // parent
																	 (HMENU)ID_LISTVIEW,        // ID
																	 g_hInst,                   // instance
																	 NULL);                     // no extra data

	if(!hwndListView)
	  return NULL;

  ListView_SetExtendedListViewStyle(hwndListView,LVS_EX_FULLROWSELECT |LVS_EX_GRIDLINES|LVS_EX_HEADERDRAGDROP  );

//	ResizeListView(hwndListView, hwndParent);


	return hwndListView;
}

/******************************************************************************

   PositionHeader

   this needs to be called when the ListView is created, resized, the view is 
   changed or a WM_SYSPARAMETERCHANGE message is recieved

******************************************************************************/

void PositionHeader(HWND hwndListView)
{
	HWND  hwndHeader = GetWindow(hwndListView, GW_CHILD);
	DWORD dwStyle = GetWindowLong(hwndListView, GWL_STYLE);

	/*To ensure that the first item will be visible, create the control without 
	the LVS_NOSCROLL style and then add it here*/
	//dwStyle |= LVS_NOSCROLL;
	//SetWindowLong(hwndListView, GWL_STYLE, dwStyle);

	//only do this if we are in report view and were able to get the header hWnd
	if(((dwStyle & LVS_TYPEMASK) == LVS_REPORT) && hwndHeader)
	{
		RECT        rc;
		HD_LAYOUT   hdLayout;
		WINDOWPOS   wpos;

		GetClientRect(hwndListView, &rc);
		hdLayout.prc = &rc;
		hdLayout.pwpos = &wpos;

		Header_Layout(hwndHeader, &hdLayout);

		SetWindowPos(  hwndHeader, 
									wpos.hwndInsertAfter, 
									wpos.x, 
									wpos.y,
									wpos.cx, 
									wpos.cy, 
									wpos.flags | SWP_SHOWWINDOW);

		ListView_EnsureVisible(hwndListView, 0, FALSE);
	}
}

/******************************************************************************

   ResizeListView

******************************************************************************/

void ResizeListView(HWND hwndListView, HWND hwndParent)
{
	RECT  rc;

	GetClientRect(hwndParent, &rc);

	MoveWindow( hwndListView, 
							rc.left,
							205,//rc.top,
							rc.right - rc.left,
							-200+rc.bottom - rc.top,
							TRUE);

	//only call this if we want the LVS_NOSCROLL style
//	PositionHeader(hwndListView);
}

/******************************************************************************

   InsertListViewItems

******************************************************************************/

BOOL InsertListViewItems(HWND hwndListView,int count)
{
	int x,a_width,h_width;

  ShowWindow(hwndListView,SW_HIDE);
	//empty the list
	ListView_DeleteAllItems(hwndListView);

	//set the number of items in the list
	ListView_SetItemCount(hwndListView, count);
	for (x=0;x<ncols;x++)
	{
    ListView_SetColumnWidth(hwndListView,x,LVSCW_AUTOSIZE );
		a_width=ListView_GetColumnWidth(hwndListView,x);
    ListView_SetColumnWidth(hwndListView,x,LVSCW_AUTOSIZE_USEHEADER  );
		h_width=ListView_GetColumnWidth(hwndListView,x);
		if (a_width>h_width)
      ListView_SetColumnWidth(hwndListView,x,LVSCW_AUTOSIZE );
	}
  ShowWindow(hwndListView,SW_SHOWNORMAL);
	return TRUE;
}

void ClearList(HWND hwndListView)
{
	int i;

	//empty the list
	ListView_DeleteAllItems(hwndListView);
	for (i=0;i<last_ncols;i++)
		ListView_DeleteColumn(hwndListView,0);
	last_ncols=0;
}

/******************************************************************************

   InitListView

******************************************************************************/

BOOL InitListView(HWND hwndListView)
{
	LVCOLUMN   lvColumn;
	int        i,bufsize;

  last_ncols=ncols;
	//initialize the columns
	lvColumn.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM ;
	lvColumn.fmt = LVCFMT_LEFT;
	lvColumn.cx = 120;
  recordlength=0;
	bufsize=0;
	if (buffer!=NULL)
		free(buffer);
	for(i = 0; i < ncols; i++)
	{
		lvColumn.pszText = cols[i].col_name;
		ListView_InsertColumn(hwndListView, i, &lvColumn);
		recordlength+=cols[i].col_size;
    if (cols[i].col_size>bufsize)
			bufsize=cols[i].col_size;
	}
	if (bufsize)
    buffer=(char*)malloc(bufsize);
	else
		return FALSE;
	InsertListViewItems(hwndListView,nrows);

	return TRUE;
}

/**************************************************************************

   ListViewNotify()

**************************************************************************/

LRESULT ListViewNotify(HWND hWnd, LPARAM lParam)
{
	LPNMHDR  lpnmh = (LPNMHDR) lParam;
	HWND     hwndListView = GetDlgItem(hWnd, ID_LISTVIEW);

	switch(lpnmh->code)
	{
		case LVN_GETDISPINFO:
			{
				LV_DISPINFO *lpdi = (LV_DISPINFO *)lParam;
				DWORD bytes_read;

				if(lpdi->item.mask & LVIF_TEXT)
				{
					int x,offset=0;

          for (x=0;x<lpdi->item.iSubItem;x++)
						offset+=cols[x].col_size;
					SetFilePointer(f,recordlength*lpdi->item.iItem+offset,0,FILE_BEGIN);
					ReadFile(f,buffer,cols[lpdi->item.iSubItem].col_size,&bytes_read,NULL);
					lpdi->item.pszText=buffer;
				}
			}
	    return 0;

		case LVN_ODCACHEHINT:
			{
				LPNMLVCACHEHINT   lpCacheHint = (LPNMLVCACHEHINT)lParam;
				/*
				This sample doesn't use this notification, but this is sent when the 
				ListView is about to ask for a range of items. On this notification, 
				you should load the specified items into your local cache. It is still 
				possible to get an LVN_GETDISPINFO for an item that has not been cached, 
				therefore, your application must take into account the chance of this 
				occurring.
				*/
			}
			return 0;

		case LVN_ODFINDITEM:
			{
				LPNMLVFINDITEM lpFindItem = (LPNMLVFINDITEM)lParam;
				/*
				This sample doesn't use this notification, but this is sent when the 
				ListView needs a particular item. Return -1 if the item is not found.
				*/
			}
			return 0;
	}

	return 0;
}

