#include "jbsql.h"

int linepos=0;

void RemoveLines(TextWidget *text,int nlines)
{
  int x;

  for (x=text->lines-1;x>text->lines-1-nlines;x--)
  {
    free(text->text[x]);
  }
  text->lines-=nlines;
  if (text->cur_line>text->lines)
    text->cur_line=text->lines;
  if (text->top+text->height>text->lines)
    text->top=0;
}

void SyncTextPosition(TextWidget *text)
{
  int x;

  text->cur_line=text->lines;
  text->cur_pos=strlen(text->text[text->cur_line-1]);
  for (x=strlen(command)-1;x>=linepos;x--)
  {
    if (command[x]=='\n')
    {
      text->cur_line--;
      text->cur_pos=strlen(text->text[text->cur_line-1]);
    }
    else
    {
      text->cur_pos--;
    }
  }
}

__inline void AppendLineBuffer(char *str)
{
  int old_command_size=cur_command_size;

  ReallocCommand(strlen(str));
  //command=realloc(command,strlen(command)+strlen(str)+1);*/
  //if (linepos<(int)strlen(command)) /* Insert some text */
  if (linepos<old_command_size-1)
  {
    memmove(&command[linepos+strlen(str)],&command[linepos],strlen(&command[linepos])+1);
    memcpy(&command[linepos],str,strlen(str));
  }
  else /* Append some text */
  {
    //strcat(command,str);
    memcpy(&command[old_command_size-1],str,cur_command_size-old_command_size+1);
  }
  linepos+=cur_command_size-old_command_size;
}

int DeleteLineBufferCharacter(TextWidget *text) /* Backspace */
{
  char delchar;
  int x,nlines_before=1,nlines_after=1,redrawpos=0;

  if (strlen(command) && linepos)
  {
    delchar=command[linepos-1];
    if (linepos<(int)strlen(command)) /* Remove a character in the middle */
    {
      memmove(&command[linepos-1],&command[linepos],strlen(&command[linepos])+1);
    }
    else /* Remove a character at the end */
      command[strlen(command)-1]='\0';
    ReallocCommand(-1);
/*    command=realloc(command,strlen(command)+1);*/
    linepos--;
    if (delchar=='\n')
    {
      for (x=0;x<linepos;x++)
      {
        if (command[x]=='\n')
          nlines_before++;
      }
      for (x=linepos;x<(int)strlen(command);x++)
      {
        if (command[x]=='\n')
          nlines_after++;
      }
      for (x=linepos-1;x>=0;x--)
      {
        if (command[x]=='\n')
        {
          redrawpos=x+1;
          break;
        }
      }
      if (text->termout)
      {
        RemoveLines(text,nlines_after);
        RedrawCommand(text,redrawpos,nlines_before);
        SyncTextPosition(text);
        UpdateCaretPosition(text,1);
      }
      nlines--;
      return 0;
    }
    return 1;
  }
  return 0;
}

int DeleteNextLineBufferCharacter(TextWidget *text) /* Delete */
{
  char delchar;
  int x,nlines_before=1,nlines_after=1,redrawpos=0;

  if (strlen(command) && linepos<(int)strlen(command))
  {
    delchar=command[linepos];
    memmove(&command[linepos],&command[linepos+1],strlen(&command[linepos+1])+1);
    ReallocCommand(1);
/*    command=realloc(command,strlen(command)+1);*/
    if (delchar=='\n')
    {
      for (x=0;x<linepos;x++)
      {
        if (command[x]=='\n')
          nlines_before++;
      }
      for (x=linepos;x<(int)strlen(command);x++)
      {
        if (command[x]=='\n')
          nlines_after++;
      }
      for (x=linepos-1;x>=0;x--)
      {
        if (command[x]=='\n')
        {
          redrawpos=x+1;
          break;
        }
      }
      if (text->termout)
      {
        RemoveLines(text,nlines_after);
        RedrawCommand(text,redrawpos,nlines_before);
        SyncTextPosition(text);
        UpdateCaretPosition(text,1);
      }
      nlines--;
      return 0;
    }
    return 1;
  }
  return 0;
}

void InsertTab(TextWidget *text)
{
  int ntabs,pos=linepos,x;
  char tabstr[10];

  for (x=linepos-1;x>=0;x--)
  {
    if (command[x]=='\n')
    {
      pos=linepos-x;
      break;
    }
  }
  ntabs=8-pos%8;
  sprintf(tabstr,"%*s",ntabs,"");
  AppendLineBuffer(tabstr);
  if (text->termout)
    WPrintf(tabstr);
}

__inline void InsertNewLine(TextWidget *text)
{
  int x,nlines_before=1,nlines_after=1,redrawpos=0;

  if (command[0]=='\0')
  {
    nlines=1;
    if (text->termout)
    {
      ShowPrompt(text);
    }
  }
  else if (!allow_blank_lines && command[strlen(command)-1]=='\n' && !text->no_process)
  {
    ClearCommand(text,1,0);
  }
  else if (linepos<(int)strlen(command))
  {
    allow_blank_lines=0;
    for (x=0;x<linepos;x++)
    {
      if (command[x]=='\n')
        nlines_before++;
    }
    for (x=linepos;x<(int)strlen(command);x++)
    {
      if (command[x]=='\n')
        nlines_after++;
    }
    for (x=linepos-1;x>=0;x--)
    {
      if (command[x]=='\n')
      {
        redrawpos=x+1;
        break;
      }
    }
    nlines++;
    if (text->termout)
    {
      AppendLineBuffer("\n");
      RemoveLines(text,nlines_after);
      RedrawCommand(text,redrawpos,nlines_before);
      SyncTextPosition(text);
      UpdateCaretPosition(text,1);
    }
  }
  else
  {
    AppendLineBuffer("\n");
    nlines++;
    if (text->termout)
      RedrawCommand(text,linepos,nlines);
  }
}

void InsertLineBuffer(TextWidget *text,char *fmt,...)
{
  va_list param;
  char *str;
  int x,lastx=0,slen;

  va_start(param,fmt);
  str=myvsprintf(fmt,param);
  va_end(param);

  slen=strlen(str);
  for (x=0;x<(int)slen;x++)
  {
    switch (str[x])
    {
      case '\n':
        str[x]='\0';
        AppendLineBuffer(&str[lastx]);
        if (text->termout)
          WPrintf("%s\n",&str[lastx]);
        if (text->scanf)
          text->scanf=0;
        else
        {
          if (!text->no_process && linepos && linepos==(int)strlen(command))
          {
            switch (ProcessCommand(text))
            {
              case 0:
                InsertNewLine(text);
                break;
                break;
            }
          }
          else
            InsertNewLine(text);
        }
        lastx=x+1;
        break;
      case '\r':
        str[x]=' ';
        break;
      case '\t':
        str[x]='\0';
        AppendLineBuffer(&str[lastx]);
        InsertTab(text);
        lastx=x+1;
        break;
      case 8:
        if (DeleteLineBufferCharacter(text))
        {
          if (text->termout)
            WPrintf("%c",8);
        }
        lastx=x+1;
        break;
      default:
        break;
    }
  }
  if (lastx<(int)slen)
  {
    AppendLineBuffer(&str[lastx]);
    if (text->termout)
      WPrintfUP(&str[lastx]);
  }
  free(str);
}

void InsertUnparsedLineBuffer(TextWidget *text,char *str)
{
  AppendLineBuffer(str);
  if (text->termout)
    WPrintfUP(str);
}

void InsertNewLineBuffer(TextWidget *text)
{
  if (text->termout)
    WPrintfNL();
  if (text->scanf)
    text->scanf=0;
  else
  {
    if (!text->no_process && linepos && linepos==(int)strlen(command))
    {
      switch (ProcessCommand(text))
      {
        case 0:
          InsertNewLine(text);
          break;
      }
    }
    else
      InsertNewLine(text);
  }
}

void MoveCursorLeft(TextWidget *text,int next_word)
{
  if (next_word)
  {
    while (linepos && strchr(MOVE_DELIMITERS,command[linepos-1]))
    {
      linepos--;
      if (command[linepos]=='\n')
      {
        text->cur_line--;
        text->cur_pos=strlen(text->text[text->cur_line-1]);
      }
      else
        text->cur_pos--;
    }
    while (linepos && !strchr(MOVE_DELIMITERS,command[linepos-1]))
    {
      linepos--;
      if (command[linepos]=='\n')
      {
        text->cur_line--;
        text->cur_pos=strlen(text->text[text->cur_line-1]);
      }
      else
        text->cur_pos--;
    }
  }
  else
  {
    if (linepos)
    {
      linepos--;
      if (command[linepos]=='\n')
      {
        text->cur_line--;
        text->cur_pos=strlen(text->text[text->cur_line-1]);
      }
      else
        text->cur_pos--;
    }
  }
  UpdateCaretPosition(text,1);
}

void MoveCursorRight(TextWidget *text,int next_word)
{
  if (next_word)
  {
    while (linepos<(int)strlen(command) && !strchr(MOVE_DELIMITERS,command[linepos]))
    {
      if (command[linepos]=='\n')
      {
        text->cur_line++;
        text->cur_pos=strlen(STR_OPTION(PROMPT))+1;
      }
      else
        text->cur_pos++;
      linepos++;
    }
    while (linepos<(int)strlen(command) && strchr(MOVE_DELIMITERS,command[linepos]))
    {
      if (command[linepos]=='\n')
      {
        text->cur_line++;
        text->cur_pos=strlen(STR_OPTION(PROMPT))+1;
      }
      else
        text->cur_pos++;
      linepos++;
    }
  }
  else
  {
    if (linepos<(int)strlen(command))
    {
      if (command[linepos]=='\n')
      {
        text->cur_line++;
        text->cur_pos=strlen(STR_OPTION(PROMPT))+1;
      }
      else
        text->cur_pos++;
      linepos++;
    }
  }
  UpdateCaretPosition(text,1);
}

void MoveCursorHome(TextWidget *text,int full)
{
  if (full)
  {
    linepos=0;
  }
  else
  {
    while (linepos && command[linepos-1]!='\n')
      linepos--;
  }
  SyncTextPosition(text);
  UpdateCaretPosition(text,1);
}

void MoveCursorEnd(TextWidget *text,int full)
{
  if (full)
  {
    linepos=strlen(command);
  }
  else
  {
    while (linepos<(int)strlen(command) && command[linepos]!='\n')
      linepos++;
  }
  SyncTextPosition(text);
  UpdateCaretPosition(text,1);
}

void MoveCursorUp(TextWidget *text)
{
  int x,nlpos1=0,nlpos2=0,xpos=0;

  for (x=linepos-1;x>=0;x--)
  {
    if (command[x]=='\n')
    {
      nlpos1=x;
      break;
    }
  }
  if (!nlpos1)
    return;
  xpos=linepos-nlpos1-1;
  for (x=nlpos1-1;x>=0;x--)
  {
    if (command[x]=='\n')
    {
      nlpos2=x+1;
      break;
    }
  }
  if (nlpos2+xpos>nlpos1)
    linepos=nlpos1;
  else
    linepos=nlpos2+xpos;
  SyncTextPosition(text);
  UpdateCaretPosition(text,1);
}

void MoveCursorDown(TextWidget *text)
{
  int x,nlpos1=-1,nlpos2,xpos=0;

  for (x=linepos-1;x>=0;x--)
  {
    if (command[x]=='\n')
    {
      nlpos1=x;
      break;
    }
  }
  xpos=linepos-nlpos1-1;
  nlpos2=strlen(command);
  for (x=linepos;x<=(int)strlen(command);x++)
  {
    if (command[x]=='\n')
    {
      nlpos2=x+1;
      break;
    }
  }
  linepos=nlpos2;
  for (x=nlpos2;x<MIN((int)strlen(command),nlpos2+xpos);x++)
  {
    if (command[x]=='\n')
      break;
    linepos=x+1;
  }
  SyncTextPosition(text);
  UpdateCaretPosition(text,1);
}

void AppendLine(TextWidget *text,char *str)
{
  text->text[text->cur_line-1]=realloc(text->text[text->cur_line-1],strlen(text->text[text->cur_line-1])+strlen(str)+1);
  if (text->cur_pos<(int)strlen(text->text[text->cur_line-1])) /* Insert some text */
  {
    memmove(&text->text[text->cur_line-1][text->cur_pos+strlen(str)],&text->text[text->cur_line-1][text->cur_pos],strlen(&text->text[text->cur_line-1][text->cur_pos])+1);
    if (text->scanf==2)
      memset(&text->text[text->cur_line-1][text->cur_pos],'*',strlen(str));
    else
      memcpy(&text->text[text->cur_line-1][text->cur_pos],str,strlen(str));
  }
  else /* Append some text */
  {
    if (text->scanf==2)
    {
      text->text[text->cur_line-1][strlen(text->text[text->cur_line-1])+strlen(str)]='\0';
      memset(&text->text[text->cur_line-1][strlen(text->text[text->cur_line-1])],'*',strlen(str));
    }
    else
      strcat(text->text[text->cur_line-1],str);
  }
  text->cur_pos+=strlen(str);
}

void InsertLine(TextWidget *text)
{
  text->lines++;
  text->cur_line++;
  if (text->lines>(int)INT_OPTION(MAXLINES))
  {
    text->lines=INT_OPTION(MAXLINES);
    free(text->text[0]);
    memmove(&(text->text[0]),&(text->text[1]),(INT_OPTION(MAXLINES)-1)*sizeof(char*));
  }
  else
  {
    text->text=(char**)realloc(text->text,text->lines*sizeof(char*));
  }
  if (text->cur_line>(int)INT_OPTION(MAXLINES))
    text->cur_line=INT_OPTION(MAXLINES);
  text->text[text->lines-1]=strdup("");
  text->cur_pos=0;
  text->left=0;
}

void ClearLine(TextWidget *text)
{
  free(text->text[text->lines-1]);
  text->text[text->lines-1]=strdup("");
  text->cur_pos=0;
  text->left=0;
}

void DeleteCharacter(TextWidget *text) /* Backspace */
{
  if (strlen(text->text[text->cur_line-1]) && text->cur_pos)
  {
    if (text->cur_pos<(int)strlen(text->text[text->cur_line-1])) /* Remove a character in the middle */
    {
      memmove(&text->text[text->cur_line-1][text->cur_pos-1],&text->text[text->cur_line-1][text->cur_pos],strlen(&text->text[text->cur_line-1][text->cur_pos])+1);
    }
    else /* Remove at the end */
      text->text[text->cur_line-1][strlen(text->text[text->cur_line-1])-1]='\0';
    text->text[text->cur_line-1]=realloc(text->text[text->cur_line-1],strlen(text->text[text->cur_line-1])+1);
    text->cur_pos--;
  }
}

void DeleteNextCharacter(TextWidget *text) /* Delete */
{
  if (strlen(text->text[text->cur_line-1]) && text->cur_pos<(int)strlen(text->text[text->cur_line-1]))
  {
    memmove(&text->text[text->cur_line-1][text->cur_pos],&text->text[text->cur_line-1][text->cur_pos+1],strlen(&text->text[text->cur_line-1][text->cur_pos+1])+1);
    text->text[text->cur_line-1]=realloc(text->text[text->cur_line-1],strlen(text->text[text->cur_line-1])+1);
  }
  InvalidateRect(hwndTextBox, NULL, TRUE);
  UpdateWindow(hwndTextBox);
}
