#include "jbsql.h"

SQLHint sqlhints[]={
{"None",0,0,"Let the optimizer decide the best plan"},
{"ALL_ROWS",0,0,"The ALL_ROWS hint explicitly chooses the cost-based approach to optimize a statement block with a goal of best throughput (that is, minimum total resource consumption)"},
{"FIRST_ROWS",0,0,"The hints FIRST_ROWS(n) (where n is any positive integer) or FIRST_ROWS instruct Oracle to optimize an individual SQL statement for fast response. FIRST_ROWS(n) affords greater precision, because it instructs Oracle to choose the plan that returns the first n rows most efficiently. The FIRST_ROWS hint, which optimizes for the best plan to return the first single row, is retained for backward compatibility and plan stability"},
{"CHOOSE",0,0,"The CHOOSE hint causes the optimizer to choose between the rule-based and cost-based approaches for a SQL statement. The optimizer bases its selection on the presence of statistics for the tables accessed by the statement. If the data dictionary has statistics for at least one of these tables, then the optimizer uses the cost-based approach and optimizes with the goal of best throughput. If the data dictionary does not have statistics for these tables, then it uses the rule-based approach."},
{"RULE",0,0,"Forces Oracle to use rule based optimizer"},
{"FULL",1,0,"The FULL hint explicitly chooses a full table scan for the specified table"},
{"ROWID",1,0,"The ROWID hint explicitly chooses a table scan by rowid for the specified table"},
{"CLUSTER",1,0,"The CLUSTER hint explicitly chooses a cluster scan to access the specified table. It applies only to clustered objects."},
{"HASH",1,0,"The HASH hint explicitly chooses a hash scan to access the specified table. It applies only to tables stored in a cluster"},
{"INDEX",1,2,"The INDEX hint explicitly chooses an index scan for the specified table. You can use the INDEX hint for domain, B-tree, bitmap, and bitmap join indexes"},
{"INDEX_ASC",1,2,"The INDEX_ASC hint explicitly chooses an index scan for the specified table. If the statement uses an index range scan, then Oracle scans the index entries in ascending order of their indexed values"},
{"INDEX_COMBINE",1,2,"The INDEX_COMBINE hint explicitly chooses a bitmap access path for the table. If no indexes are given as arguments for the INDEX_COMBINE hint, then the optimizer uses whatever Boolean combination of bitmap indexes has the best cost estimate for the table. If certain indexes are given as arguments, then the optimizer tries to use some Boolean combination of those particular bitmap indexes"},
{"INDEX_JOIN",1,2,"The INDEX_JOIN hint explicitly instructs the optimizer to use an index join as an access path. For the hint to have a positive effect, a sufficiently small number of indexes must exist that contain all the columns required to resolve the query"},
{"INDEX_DESC",1,2,"The INDEX_DESC hint explicitly chooses an index scan for the specified table. If the statement uses an index range scan, then Oracle scans the index entries in descending order of their indexed values. In a partitioned index, the results are in descending order within each partition"},
{"INDEX_FFS",1,2,"The INDEX_FFS hint causes a fast full index scan to be performed rather than a full table scan"},
{"NO_INDEX",1,2,"The NO_INDEX hint explicitly disallows a set of indexes for the specified table"},
{"AND_EQUAL",1,2,"The AND_EQUAL hint explicitly chooses an execution plan that uses an access path that merges the scans on several single-column indexes"},
{"USE_CONCAT",0,0,"The USE_CONCAT hint forces combined OR conditions in the WHERE clause of a query to be transformed into a compound query using the UNION ALL set operator. Generally, this transformation occurs only if the cost of the query using the concatenations is cheaper than the cost without them."},
{"NO_EXPAND",0,0,"The NO_EXPAND hint prevents the cost-based optimizer from considering OR-expansion for queries having OR conditions or IN-lists in the WHERE clause. Usually, the optimizer considers using OR expansion and uses this method if it decides that the cost is lower than not using it"},
{"REWRITE",0,0,"The REWRITE hint forces the cost-based optimizer to rewrite a query in terms of materialized views, when possible, without cost consideration. Use the REWRITE hint with or without a view list. If you use REWRITE with a view list and the list contains an eligible materialized view, then Oracle uses that view regardless of its cost."},
{"NOREWRITE",0,0,"The NOREWRITE hint disables query rewrite for the query block, overriding the setting of the parameter QUERY_REWRITE_ENABLED. Use the NOREWRITE hint on any query block of a request"},
{"MERGE",1,0,"If a view's query contains a GROUP BY clause or DISTINCT operator in the SELECT list, then the optimizer can merge the view's query into the accessing statement only if complex view merging is enabled. Complex merging can also be used to merge an IN subquery into the accessing statement if the subquery is uncorrelated"},
{"NO_MERGE",1,0,"The NO_MERGE hint causes Oracle not to merge mergeable views"},
{"STAR_TRANSFORMATION",0,0,"The STAR_TRANSFORMATION hint makes the optimizer use the best plan in which the transformation has been used. Without the hint, the optimizer could make a cost-based decision to use the best plan generated without the transformation, instead of the best plan for the transformed query"},
{"FACT",1,0,"The FACT hint is used in the context of the star transformation to indicate to the transformation that the hinted table should be considered as a fact table."},
{"NO_FACT",1,0,"The NO_FACT hint is used in the context of the star transformation to indicate to the transformation that the hinted table should not be considered as a fact table."},
{"ORDERED",0,0,"The ORDERED hint causes Oracle to join tables in the order in which they appear in the FROM clause"},
{"STAR",0,0,"The STAR hint forces a star query plan to be used, if possible. A star plan has the largest table in the query last in the join order and joins it with a nested loops join on a concatenated index. The STAR hint applies when there are at least three tables, the large table's concatenated index has at least three columns, and there are no conflicting access or join method hints. The optimizer also considers different permutations of the small tables."},
{"USE_NL",2,0,"The USE_NL hint causes Oracle to join each specified table to another row source with a nested loops join using the specified table as the inner table"},
{"USE_MERGE",2,0,"The USE_MERGE hint causes Oracle to join each specified table with another row source with a sort-merge join"},
{"USE_HASH",2,0,"The USE_HASH hint causes Oracle to join each specified table with another row source with a hash join"},
{"DRIVING_SITE",1,0,"The DRIVING_SITE hint forces query execution to be done at a different site than that selected by Oracle. This hint can be used with either rule-based or cost-based optimization"},
{"LEADING",1,0,"The LEADING hint causes Oracle to use the specified table as the first table in the join order"},
};

int nhints=sizeof(sqlhints)/sizeof(SQLHint);